@@ -1,28 +1,13 @@
 import contextlib
-import time
 
 import pytest
-from _pytest.main import (
-    EXIT_OK,
-    EXIT_TESTSFAILED,
-    EXIT_INTERRUPTED,
-    EXIT_USAGEERROR,
-    EXIT_NOTESTSCOLLECTED,
-)
+import six
 from _pytest.runner import runtestprotocol
+from _pytest.junitxml import _NodeReporter
 
 from pytest_sherlock.binary_tree_search import Root as BTSRoot
 
 
-PYTEST_EXIT_CODES = (
-    EXIT_OK,
-    EXIT_TESTSFAILED,
-    EXIT_INTERRUPTED,
-    EXIT_USAGEERROR,
-    EXIT_NOTESTSCOLLECTED,
-)
-
-
 def build_summary_stats_line(stats):
     keys = [
         "failed", "passed", "skipped", "deselected", "xfailed", "xpassed", "warnings", "error",
@@ -85,13 +70,47 @@ def _remove_failed_setup_state_from_session(item):
     return True
 
 
+class Bucket(object):
+    def __init__(self, items):
+        self.items = items
+        self.__current = 0
+        self.failed_report = None
+
+    def __len__(self):
+        return len(self.items)
+
+    def __iter__(self):
+        return self
+
+    def __getitem__(self, item):
+        return self.items[item]
+
+    def __repr__(self):
+        return str(self.items)
+
+    def __str__(self):
+        return "<Bucket items={}>".format(len(self.items))
+
+    def next(self):
+        if self.__current < len(self.items):
+            item = self.items[self.__current]
+            self.__current += 1
+            return item
+
+        self.__current = 0
+        raise StopIteration
+
+
 class Collection(object):
 
     def __init__(self, items):
         self.items = items
         self.test_func = None
+        self._bts = BTSRoot()
+        self.__current_root = None
+        self.__last = None
 
-    def needed_tests(self, test_name):
+    def _needed_tests(self, test_name):
         tests = []
         for test_func in self.items:
             if test_func.name == test_name or test_func.nodeid == test_name:
@@ -113,11 +132,45 @@ class Collection(object):
         )
         return tests
 
+    def __len__(self):
+        return len(self._bts)
+
+    def __iter__(self):
+        return self
+
+    def next(self):
+        if self.__last is not None:
+            self._set_current_status(not bool(self.__last.failed_report))
+            self.__last = self._get_current_tests()
+            return self.__last
+
+        self.__last = self._get_current_tests()
+        return self.__last
+
+    def prepare(self, test_name):
+        items = self._needed_tests(test_name)
+        self._bts.insert(items)
+        self.__current_root = self._bts.root
+
+    def _set_current_status(self, status):
+        if status is True:
+            self.__current_root = self.__current_root.right
+        else:
+            self.__current_root = self.__current_root.left
+
+    def _get_current_tests(self):
+        if self.__current_root is None:
+            return
+        if self.__current_root.left is not None:
+            return Bucket(self.__current_root.left.items)
+        return Bucket(self.__current_root.items)
+
 
 class Sherlock(object):
     def __init__(self, config):
         self.config = config
-        self.bts_root = BTSRoot()
+        # TODO add tests
+        self.collection = None
         self._tw = None
         self._reporter = None
         self._coupled = None
@@ -134,17 +187,16 @@ class Sherlock(object):
             self._tw = self.reporter.writer
         return self._tw
 
-    def write_step(self, step):
+    def write_step(self, step, maximum):
         self.reporter.ensure_newline()
-        # self.terminal.sep(sep, title, **markup)
-        # TODO need add process like: Step: [1 of 12]
-        self.terminal.write("Step #{}:".format(step), yellow=True, bold=True)
+        message = "Step [{} of {}]:".format(step, maximum)
+        self.terminal.sep('_', message, yellow=True, bold=True)
 
     @contextlib.contextmanager
     def log(self, item):
-        self.reporter.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)
-        yield self.reporter.pytest_runtest_logreport
-        self.reporter.pytest_runtest_logfinish(nodeid=item.nodeid)
+        item.ihook.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)
+        yield item.ihook.pytest_runtest_logreport
+        item.ihook.pytest_runtest_logfinish(nodeid=item.nodeid, location=item.location)
 
     @staticmethod
     def refresh_state(item):
@@ -157,87 +209,99 @@ class Sherlock(object):
         self.reporter._progress_nodeids_reported = set()
         self.reporter._session.testscollected = len(collection) + 1  # current item
 
-    def summary_coupled(self):
-        # TODO port class TerminalReporter and modify
-        if self.config.option.tbstyle != "no":
-            reports = self.reporter.getreports("coupled")
-            if not reports:
-                return
-            last_report = reports[-1]
-            if self._coupled:
-                coupled_test_names = [t.nodeid for t in self._coupled + [last_report]]
-                msg = "found coupled tests: \n\t - {coupled}\n\npytest -l -vv {tests}\n".format(
-                    coupled="\n\t - ".join(coupled_test_names),
-                    tests=" ".join(coupled_test_names),
-                )
-                self.reporter.write(msg, red=True)
-
-            msg = self.reporter._getfailureheadline(last_report)
-            self.reporter.write_sep("_", msg, red=True, bold=True)
-            self.reporter._outrep_summary(last_report)
-
-    def summary_stats(self):
-        session_duration = time.time() - self.reporter._sessionstarttime
-        line, color = build_summary_stats_line(self.reporter.stats)
-        msg = "%s in %.2f seconds" % (line, session_duration)
-        markup = {color: True, 'bold': True}
-
-        if self.reporter.verbosity >= 0:
-            self.reporter.write_sep("=", msg, **markup)
-        if self.reporter.verbosity == -1:
-            self.reporter.write_line(msg, **markup)
+    def write_coupled_report(self):
+        # TODO can I get info about modified common fixtures?
+        coupled_test_names = [t.nodeid.replace("::()::", "::") for t in self._coupled]
+        common_fixtures = set.intersection(*[set(t.fixturenames) for t in self._coupled])
+        msg = (
+            "Found coupled tests:\n"
+            "{coupled}\n\n"
+            "Common fixtures:\n"
+            "{common_fixtures}\n\n"
+            "How to reproduce:\npytest -l -vv {tests}\n"
+        ).format(
+            coupled="\n".join(coupled_test_names),
+            tests=" ".join(coupled_test_names),
+            common_fixtures="\n".join(common_fixtures) if common_fixtures else ""
+        )
+        return msg
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_collection_modifyitems(self, session, config, items):
-        """
+        r"""
         called after collection has been performed, may filter or re-order
         the items in-place.
 
+        Items:
+        [
+            'tests/test_one.py::test_one',
+            'tests/test_two.py::test_two',
+            'tests/test_tree.py::test_tree',
+            'tests/test_four.py::test_four',
+            'tests/test_five.py::test_five',
+            'tests/test_six.py::test_six',
+        ]
+        Target item: 'tests/test_five.py::test_five'
+        Prepared collection (filtered by target item):
+        ____________________________________________
+        || index | test                           ||
+        --------------------------------------------
+        | 0      | 'tests/test_one.py::test_one'   |
+        | 1      | 'tests/test_two.py::test_two'   |
+        | 2      | 'tests/test_tree.py::test_tree' |
+        | 3      | 'tests/test_four.py::test_four' |
+        --------------------------------------------
+
+        Binary tree search:
+        # First step
+                            root
+                           /    \
+                     [0, 1]      [2, 3]
+        # Second step
+                    /   \        /    \
+                 [0]    [1]    [2]    [3]
+
         :param _pytest.config.Config config: pytest config object
         :param List[_pytest.nodes.Item] items: list of item objects
         """
         if config.getoption("--flaky-test"):
-            test_collection = Collection(items)
-            self.bts_root.insert(test_collection.needed_tests(config.option.flaky_test))
-            items[:] = [test_collection.test_func]
+            self.collection = Collection(items)
+            self.collection.prepare(config.option.flaky_test)
+            items[:] = [self.collection.test_func]
         yield
 
     @pytest.hookimpl(trylast=True)
     def pytest_report_collectionfinish(self, config, startdir, items):
-        # TODO add bts length, from 12 to 13 steps
-        return "Try to find coupled tests"
+        length = len(self.collection)
+        return "Try to find coupled tests in [{}-{}] steps".format(length, length + 1)
 
-    @pytest.hookimpl(hookwrapper=True)
-    def pytest_terminal_summary(self, terminalreporter):
-        if self._coupled:
-            self.summary_stats()
-        self.summary_coupled()
-        yield
-        terminalreporter.summary_failures()
-        terminalreporter.summary_errors()
-        terminalreporter.summary_warnings()
-        terminalreporter.summary_passes()
+    def patch_report(self, item, failed_report):
+        if hasattr(failed_report.longrepr, "reprcrash"):
+            message = failed_report.longrepr.reprcrash.message
+        elif isinstance(failed_report.longrepr, six.string_types):
+            message = failed_report.longrepr
+        else:
+            message = str(failed_report.longrepr)
+        failed_report.longrepr = "\n{}\n\n{}".format(self.write_coupled_report(), message)
+        node_reporter = _NodeReporter(item.nodeid, self.config._xml)
+        node_reporter.append_failure(failed_report)
+        node_reporter.finalize()
+        self.config._xml.node_reporters_ordered[:] = [node_reporter]
+        self.config._xml.stats["failure"] = 1
+        reports = self.reporter.getreports("failed")
+        last_report = reports[-1]
+        self.reporter.stats["failed"] = [last_report]
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_protocol(self, item, nextitem):
-        steps = 1
-        root = self.bts_root.root
-        while root is not None:
-            self.write_step(steps)
-            # TODO need little bit refactor BTS, add sugar
-            current_tests = root.left.value if root.left is not None else root.value
-            self.reset_progress(current_tests)
-            self.call_items(target_item=item, items=current_tests)
-            is_target_test_success = self.call_target(target_item=item)
-            if is_target_test_success:
-                root = root.right
-            else:
-                if len(current_tests) == 1:
-                    self._coupled = current_tests
-                    return False
-                root = root.left
-            steps += 1
-        return True
+        max_length = len(self.collection) + 1  # target test
+        for step, bucket in enumerate(self.collection, start=1):
+            self.write_step(step, max_length)
+            self.call(target_item=item, items=bucket)
+            if len(bucket) == 1 and bucket.failed_report:
+                self._coupled = [bucket[0], item]
+                self.patch_report(item, bucket.failed_report)
+                break
 
     def pytest_runtestloop(self, session):
         if session.testsfailed and not session.config.option.continue_on_collection_errors:
@@ -266,16 +330,25 @@ class Sherlock(object):
                     logger(report=report)
 
     def call_target(self, target_item):
-        success = []
+        failed_report = None
         with self.log(target_item) as logger:
             reports = runtestprotocol(target_item, log=False)
             for report in reports:  # 3 reports: setup, call, teardown
                 if report.failed is True:
-                    report.outcome = 'coupled'
+                    # TODO need to rewrite
+                    # report.outcome = 'coupled'
                     self.refresh_state(item=target_item)
                     logger(report=report)
-                    success.append(False)
+                    failed_report = report
                     continue
                 logger(report=report)
-                success.append(True)
-        return all(success)  # setup, call, teardown must success
+        return failed_report  # setup, call, teardown must success
+
+    def call(self, target_item, items):
+        """
+        :param target_item: current flaky test
+        :param Bucket items: bucket of tests which probably guilty in failure
+        """
+        self.reset_progress(items)
+        self.call_items(target_item=target_item, items=items)
+        items.failed_report = self.call_target(target_item=target_item)
